/*
Copyright © 2023 Philipp Wolfer <phw@uploadedlobster.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/
package scrobblerlog

import (
	"bufio"
	"encoding/csv"
	"fmt"
	"io"
	"strconv"
	"strings"
	"time"

	"go.uploadedlobster.com/scotty/internal/models"
)

type ScrobblerLog struct {
	Timezone string
	Client   string
	Listens  models.ListensList
}

func Parse(data io.Reader, includeSkipped bool) (ScrobblerLog, error) {
	result := ScrobblerLog{
		Listens: make(models.ListensList, 0),
	}

	reader := bufio.NewReader(data)
	err := ReadHeader(reader, &result)
	if err != nil {
		return result, err
	}

	tsvReader := csv.NewReader(reader)
	tsvReader.Comma = '\t'
	// Row length is often flexible
	tsvReader.FieldsPerRecord = -1

	for {
		// A row is:
		// artistName	releaseName	trackName	trackNumber	duration	rating	timestamp	recordingMbid
		row, err := tsvReader.Read()
		if err == io.EOF {
			break
		} else if err != nil {
			return result, err
		}

		// fmt.Printf("row: %v\n", row)

		// We consider only the last field (recording MBID) optional
		if len(row) < 7 {
			line, _ := tsvReader.FieldPos(0)
			return result, fmt.Errorf("invalid record in scrobblerlog line %v", line)
		}

		rating := row[5]
		if !includeSkipped && rating == "S" {
			continue
		}

		client := strings.Split(result.Client, " ")[0]
		listen, err := rowToListen(row, client)
		if err != nil {
			return result, err
		}

		result.Listens = append(result.Listens, listen)
	}

	return result, nil
}

func Write(data io.Writer, listens models.ListensList) (lastTimestamp time.Time, err error) {
	tsvWriter := csv.NewWriter(data)
	tsvWriter.Comma = '\t'

	for _, listen := range listens {
		if listen.ListenedAt.Unix() > lastTimestamp.Unix() {
			lastTimestamp = listen.ListenedAt
		}

		// A row is:
		// artistName	releaseName	trackName	trackNumber	duration	rating	timestamp	recordingMbid
		rating, ok := listen.AdditionalInfo["rockbox_rating"].(string)
		if !ok || rating == "" {
			rating = "L"
		}
		err = tsvWriter.Write([]string{
			listen.ArtistName(),
			listen.ReleaseName,
			listen.TrackName,
			strconv.Itoa(listen.TrackNumber),
			strconv.Itoa(int(listen.Duration.Seconds())),
			rating,
			strconv.Itoa(int(listen.ListenedAt.Unix())),
			string(listen.RecordingMbid),
		})
	}

	tsvWriter.Flush()
	return
}

func ReadHeader(reader *bufio.Reader, log *ScrobblerLog) error {
	// Skip header
	for i := 0; i < 3; i++ {
		line, _, err := reader.ReadLine()
		if err != nil {
			return err
		}

		if len(line) == 0 || line[0] != '#' {
			err = fmt.Errorf("unexpected header (line %v)", i)
		} else {
			text := string(line)
			if i == 0 && !strings.HasPrefix(text, "#AUDIOSCROBBLER/1") {
				err = fmt.Errorf("not a scrobbler log file")
			}

			timezone, found := strings.CutPrefix(text, "#TZ/")
			if found {
				log.Timezone = timezone
			}

			client, found := strings.CutPrefix(text, "#CLIENT/")
			if found {
				log.Client = client
			}
		}

		if err != nil {
			return err
		}
	}
	return nil
}

func WriteHeader(writer io.Writer, log *ScrobblerLog) error {
	headers := []string{
		"#AUDIOSCROBBLER/1.1\n",
		"#TZ/" + log.Timezone + "\n",
		"#CLIENT/" + log.Client + "\n",
	}
	for _, line := range headers {
		_, err := writer.Write([]byte(line))
		if err != nil {
			return err
		}
	}
	return nil
}

func rowToListen(row []string, client string) (models.Listen, error) {
	var listen models.Listen
	trackNumber, err := strconv.Atoi(row[3])
	if err != nil {
		return listen, err
	}

	duration, err := strconv.Atoi(row[4])
	if err != nil {
		return listen, err
	}

	timestamp, err := strconv.Atoi(row[6])
	if err != nil {
		return listen, err
	}

	listen = models.Listen{
		Track: models.Track{
			ArtistNames: []string{row[0]},
			ReleaseName: row[1],
			TrackName:   row[2],
			TrackNumber: trackNumber,
			Duration:    time.Duration(duration * int(time.Second)),
			AdditionalInfo: models.AdditionalInfo{
				"rockbox_rating": row[5],
				"media_player":   client,
			},
		},
		ListenedAt: time.Unix(int64(timestamp), 0),
	}

	if len(row) > 7 {
		listen.Track.RecordingMbid = models.MBID(row[7])
	}

	return listen, nil
}
